import abc

import tensorflow as tf

from . import utils


def read_conll_format_file(input_files, feature_index=0, label_index=1, **kwargs):
    all_features, all_labels = [], []

    def read_fn(features, labels):
        all_features.append(features)
        all_labels.append(labels)

    utils.read_conll_files(
        input_files,
        callback=read_fn,
        feature_index=feature_index,
        label_index=label_index,
        **kwargs)
    return all_features, all_labels


class TokenMapper:

    def __init__(self, vocab_file, unk_token='[UNK]', pad_token='[PAD]'):
        self.token2id = utils.load_vocab_file(vocab_file)
        self.id2token = {v: k for k, v in self.token2id.items()}
        assert len(self.token2id) == len(self.id2token)
        self.unk_token = unk_token
        self.pad_token = pad_token
        self.unk_id = self.token2id[unk_token]
        self.pad_id = self.token2id[pad_token]

    def encode(self, tokens, **kwargs):
        ids = [self.token2id.get(token, self.unk_id) for token in tokens]
        return ids

    def decode(self, ids, **kwargs):
        tokens = [self.id2token.get(_id, self.unk_token) for _id in ids]
        return tokens


CHINA_PEOPLE_DAILY_LABELS = {
    'O': 0,
    'B-LOC': 1,
    'I-LOC': 2,
    'B-ORG': 3,
    'I-ORG': 4,
    'B-PER': 5,
    'I-PER': 6
}


class LabelMapper:

    def __init__(self, label_map=CHINA_PEOPLE_DAILY_LABELS):
        self.label2id = label_map
        self.id2label = {v: k for k, v in self.label2id.items()}
        assert len(self.label2id) == len(self.id2label)

    def encode(self, labels, **kwargs):
        ids = [self.label2id.get(label, 0) for label in labels]
        return ids

    def decode(self, ids, **kwargs):
        labels = [self.id2label.get(_id, 'O') for _id in ids]
        return labels


class AbstractDatasetBuilder(abc.ABC):

    def build_train_dataset(self, input_files, **kwargs):
        raise NotImplementedError()

    def build_valid_dataset(self, input_files, **kwargs):
        raise NotImplementedError()

    def build_test_dataset(self, input_files, **kwargs):
        raise NotImplementedError()


class DatasetBuilder(AbstractDatasetBuilder):

    def __init__(self, token_mapper, label_mapper, token_pad_id=0, label_pad_id=0, **kwargs):
        super().__init__()
        self.token_mapper = token_mapper
        self.label_mapper = label_mapper
        self.token_pad_id = token_pad_id
        self.label_pad_id = label_pad_id

    def _read_files(self, input_files, **kwargs):
        raise NotImplementedError()

    def _build_dataset(self, features, labels, buffer_size=10000, sequence_maxlen=None, **kwargs):
        features = [self.token_mapper.encode(feature) for feature in features]
        labels = [self.label_mapper.encode(label) for label in labels]
        buffer_size = max(buffer_size, len(features))

        features = tf.ragged.constant(features, dtype=tf.int32)
        labels = tf.ragged.constant(labels, dtype=tf.int32)
        x_dataset = tf.data.Dataset.from_tensor_slices(features)
        x_dataset = x_dataset.map(lambda x: x)
        y_dataset = tf.data.Dataset.from_tensor_slices(labels)
        y_dataset = y_dataset.map(lambda y: y)
        dataset = tf.data.Dataset.zip((x_dataset, y_dataset))
        # filter sequences
        if sequence_maxlen is not None and sequence_maxlen > 0:
            dataset = dataset.filter(lambda x, y: tf.logical_and(
                tf.size(x) < sequence_maxlen, tf.size(y) < sequence_maxlen))
        # shuffle dataset
        dataset = dataset.shuffle(buffer_size=buffer_size, reshuffle_each_iteration=True)
        return dataset

    def build_train_dataset(self, input_files, batch_size=32, buffer_size=10000, sequence_maxlen=None, **kwargs):
        features, labels = self._read_files(input_files, **kwargs)
        dataset = self._build_dataset(
            features,
            labels,
            buffer_size=buffer_size,
            sequence_maxlen=sequence_maxlen,
            **kwargs)
        dataset = dataset.padded_batch(
            batch_size=batch_size,
            padded_shapes=([None], [None]),
            padding_values=(self.token_pad_id, self.label_pad_id))
        return dataset

    def build_valid_dataset(self, input_files, batch_size=32, buffer_size=10000, sequence_maxlen=None, **kwargs):
        dataset = self.build_train_dataset(
            input_files=input_files,
            batch_size=batch_size,
            buffer_size=buffer_size,
            sequence_maxlen=sequence_maxlen,
            **kwargs)
        return dataset

    def _read_predict_files(self, input_files, **kwargs):
        raise NotImplementedError()

    def _build_predict_dataset(self, features, sequence_maxlen=None, **kwargs):
        features = [self.token_mapper.encode(feature) for feature in features]
        features = tf.ragged.constant(features)
        dataset = tf.data.Dataset.from_tensor_slices(features)
        if sequence_maxlen is not None and sequence_maxlen > 0:
            dataset = dataset.filter(lambda x: tf.size(x) < sequence_maxlen)
        dataset = dataset.map(lambda x: x)  # convert to ragged tensor to normal tensor
        return dataset

    def build_test_dataset(self, input_files, batch_size=32, sequence_maxlen=None, with_labels=True, **kwargs):
        outputs = self._read_predict_files(input_files, with_labels, **kwargs)
        if with_labels:
            features, labels = outputs
        else:
            features, labels = outputs, None
        dataset = self._build_predict_dataset(
            features=features,
            sequence_maxlen=sequence_maxlen,
            **kwargs)
        dataset = dataset.padded_batch(
            batch_size=batch_size,
            padded_shapes=[None],
            padding_values=self.token_pad_id)
        if with_labels:
            return dataset, labels
        return dataset


class ChinaPeopleDailyBuilder(DatasetBuilder):

    def _read_files(self, input_files, **kwargs):
        return read_conll_format_file(input_files, **kwargs)

    def _read_predict_files(self, input_files, with_labels=True, **kwargs):
        features, labels = read_conll_format_file(input_files, **kwargs)
        if with_labels:
            return features, labels
        return features



import tensorflow as tf
import tensorflow_addons as tfa


class CRF(tf.keras.layers.Layer):

    def __init__(self,
                 units,
                 chain_initializer='orthogonal',
                 use_boundary=True,
                 boundary_initializer='zeros',
                 use_kernel=True,
                 **kwargs):
        super().__init__(**kwargs)
        self.crf = tfa.layers.CRF(
            units,
            chain_initializer=chain_initializer,
            use_boundary=use_boundary,
            boundary_initializer=boundary_initializer,
            use_kernel=use_kernel,
            **kwargs)
        self.units = units
        self.chain_kernel = self.crf.chain_kernel
        # record sequence length to compute loss
        self.sequence_length = None
        self.mask = None

    def call(self, inputs, mask=None):
        """Forward pass.

        Args:
            inputs: A [batch_size, max_seq_len, depth] tensor, inputs of CRF layer
            mask: A [batch_size, max_seq_len] boolean tensor, used to compulte sequence length in CRF layer

        Returns:
            potentials: A [batch_size, max_seq_len, units] tensor in train phase.
            sequence: A [batch_size, max_seq_len, units] tensor of decoded sequence in predict phase.
        """
        sequence, potentials, sequence_length, transitions = self.crf(inputs, mask=mask)
        # sequence_length is computed in both train and predict phase
        self.sequence_length = sequence_length
        # save mask, which is needed to compute accuracy
        self.mask = mask

        sequence = tf.cast(tf.one_hot(sequence, self.units), dtype=self.dtype)
        return tf.keras.backend.in_train_phase(potentials, sequence)

    def accuracy(self, y_true, y_pred):
        if len(tf.keras.backend.int_shape(y_true)) == 3:
            y_true = tf.argmax(y_true, axis=-1)
        y_pred, _ = tfa.text.crf_decode(y_pred, self.chain_kernel, self.sequence_length)
        y_pred = tf.cast(y_pred, dtype=y_true.dtype)
        equals = tf.cast(tf.equal(y_true, y_pred), y_true.dtype)
        if self.mask is not None:
            mask = tf.cast(self.mask, y_true.dtype)
            equals = equals * mask
            return tf.reduce_sum(equals) / tf.reduce_sum(mask)
        return tf.reduce_mean(equals)

    def neg_log_likelihood(self, y_true, y_pred):
        log_likelihood, _ = tfa.text.crf_log_likelihood(y_pred, y_true, self.sequence_length, self.chain_kernel)
        return tf.reduce_mean(-log_likelihood)



import os
import unittest

import numpy as np
import tensorflow as tf
import tensorflow_addons as tfa
from keras_crf.callbacks import EvaluateCallback
from keras_crf.crf import CRF
from keras_crf.dataset import ChinaPeopleDailyBuilder, LabelMapper, TokenMapper

token_mapper = TokenMapper('data/vocab.txt')
label_mapper = LabelMapper()

DATADIR = 'data/china-people-daily-ner-corpus'


class ModelTrainTest(unittest.TestCase):

    def _build_inputs(self):
        builder = ChinaPeopleDailyBuilder(token_mapper, label_mapper)
        train_dataset = builder.build_valid_dataset(os.path.join(DATADIR, 'example.dev'))
        # x = ['相', '比', '之', '下', '，', '青', '岛', '海', '牛', '队', '和', '广', '州', '松',
        #      '木', '队', '的', '雨', '中', '之', '战', '虽', '然', '也', '是', '0', ':', '0']
        # tokens = token_mapper.encode(x)
        # predict_inputs = tf.constant(tokens, shape=(1, len(tokens)), dtype=tf.int32)
        predict_dataset, labels = builder.build_test_dataset(os.path.join(DATADIR, 'example.test'), with_labels=True)
        return train_dataset, predict_dataset, labels

    def _build_model(self):
        sequence_input = tf.keras.layers.Input(shape=(None,), dtype=tf.int32, name='sequence_input')
        sequence_mask = tf.keras.layers.Lambda(lambda x: tf.greater(x, 0))(sequence_input)
        outputs = tf.keras.layers.Embedding(21128, 128)(sequence_input)
        outputs = tf.keras.layers.Dense(256)(outputs)
        crf = CRF(7)
        # mask is important to compute sequence length in CRF
        outputs = crf(outputs, mask=sequence_mask)
        model = tf.keras.Model(inputs=sequence_input, outputs=outputs)
        model.compile(
            loss=crf.neg_log_likelihood,
            metrics=[
                crf.accuracy
            ],
            optimizer=tf.keras.optimizers.Adam(4e-5)
        )
        model.summary()
        return model

    def test_train_model(self):
        model = self._build_model()
        dataset, predict_dataset, labels = self._build_inputs()
        model.fit(
            dataset,
            steps_per_epoch=10,
            epochs=2,
            validation_data=dataset,
            callbacks=[
                EvaluateCallback(token_mapper, label_mapper, os.path.join(DATADIR, 'example.dev')),
            ])

        preds = model.predict(predict_dataset, steps=1)
        for idx, p in enumerate(preds):
            # print(p)
            example = np.argmax(p, axis=-1).tolist()[:len(labels[idx])]
            print(labels[idx])
            print(label_mapper.decode(example))
            print()


if __name__ == "__main__":
    unittest.main()



import logging
import os


def read_files(input_files, callback=None, **kwargs):
    if isinstance(input_files, str):
        input_files = [input_files]
    for f in input_files:
        if not os.path.exists(f):
            logging.warning('File %s does not exist.', f)
            continue
        logging.info('Starting to read file %s', f)
        with open(f, mode='rt', encoding=kwargs.get('encoding', 'utf-8')) as fin:
            for line in fin:
                line = line.rstrip('\n')
                if callback:
                    callback(line)
        logging.info('Finished to read file %s', f)


def read_conll_files(input_files, callback=None, **kwargs):
    if isinstance(input_files, str):
        input_files = [input_files]
    feature_index, label_index = kwargs.get('feature_index', 0), kwargs.get('label_index', 1)
    for f in input_files:
        if not os.path.exists(f):
            logging.warning('File %s does not exist.', f)
            continue
        logging.info('Starting to read file %s', f)
        with open(f, mode='rt', encoding=kwargs.get('encoding', 'utf-8')) as fin:
            lines = fin.read().splitlines()
        features, labels = [], []
        for line in lines:
            parts = line.split(' ')
            if len(parts) == 1:
                if callback:
                    callback(features, labels)
                features, labels = [], []
            else:
                features.append(parts[feature_index])
                labels.append(parts[label_index])
        logging.info('Finished to read file %s', f)


def load_vocab_file(vocab_file, **kwargs):
    vocabs = {}
    lino = 0
    with open(vocab_file, mode='rt', encoding='utf8') as fin:
        for line in fin:
            v = line.rstrip('\n')
            vocabs[v] = lino
            lino += 1
    return vocabs



import unittest

import tensorflow as tf
import tensorflow_addons as tfa
from keras_crf.crf import CRF


class ModelBuildTest(unittest.TestCase):

    def _build_model(self):
        sequence_input = tf.keras.layers.Input(shape=(None,), dtype=tf.int32, name='sequence_input')
        sequence_mask = tf.keras.layers.Lambda(lambda x: tf.greater(x, 0))(sequence_input)
        outputs = tf.keras.layers.Embedding(21128, 128)(sequence_input)
        outputs = tf.keras.layers.Dense(256)(outputs)
        crf = CRF(7)
        # mask is important to compute sequence length in CRF
        outputs = crf(outputs, mask=sequence_mask)
        model = tf.keras.Model(inputs=sequence_input, outputs=outputs)
        model.compile(
            loss=crf.neg_log_likelihood,
            metrics=[crf.accuracy],
            optimizer=tf.keras.optimizers.Adam(4e-5)
        )
        return model

    def test_build_model(self):
        model = self._build_model()
        model.summary()


if __name__ == "__main__":
    unittest.main()



import numpy as np
import tensorflow as tf
from seqeval.metrics import classification_report

from .dataset import read_conll_format_file


class EvaluateCallback(tf.keras.callbacks.Callback):

    def __init__(self, token_mapper, label_mapper, valid_files, batch_size=32, sequence_maxlen=512):
        self.token_mapper = token_mapper
        self.laber_mapper = label_mapper
        self.batch_size = batch_size
        self.sequence_maxlen = sequence_maxlen

        self.valid_features, self.valid_labels = [], []
        features, labels = read_conll_format_file(valid_files)
        for f, l in zip(features, labels):
            if len(f) > sequence_maxlen:
                continue
            f = self.token_mapper.encode(f)
            # l = self.laber_mapper.encode(l)
            self.valid_features.append(f)
            self.valid_labels.append(l)

    def _build_dataset(self, features, batch_size=32):
        maxlen = max([len(x) for x in features])
        features = [x + [0] * (maxlen - len(x)) for x in features]
        x = tf.constant(features)
        x = tf.data.Dataset.from_tensor_slices(x)
        x = x.batch(batch_size)
        return x

    def on_epoch_end(self, epoch, logs=None):
        dataset = self._build_dataset(self.valid_features, batch_size=self.batch_size)
        # shape: [batch_size, max_seq_len, num_class]
        predictions = self.model.predict(dataset)
        # shape: [batch_size, max_seq_len]
        predictions = np.argmax(predictions, axis=-1)
        y_true = self.valid_labels
        y_pred = []
        for _pred, _true in zip(predictions, y_true):
            # remove paddings
            _pred = _pred[:len(_true)]
            tags = self.laber_mapper.decode(_pred.tolist())
            y_pred.append(tags)
        report = classification_report(y_true, y_pred)
        print()
        print('reports of epoch {}\n'.format(epoch + 1))
        print(report)



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="keras_crf",
    version="0.1.0",
    description="A more elegant and convenient CRF built on tensorflow-addons.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/luozhouyang/keras-crf",
    author="ZhouYang Luo",
    author_email="zhouyang.luo@gmail.com",
    packages=setuptools.find_packages(),
    include_package_data=True,
    install_requires=[
        'tensorflow',
        'tensorflow-addons',
        'seqeval'
    ],
    extras_require={
        
    },
    license="Apache Software License",
    classifiers=(
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
    )
)